"""
Gracefully and lightweight wrapper to deal with QIWI API
It's an open-source project so you always can improve the quality of code/API by
adding something of your own...
Easy to integrate to Telegram bot, which was written on aiogram or another async/sync library.

"""
from __future__ import annotations

import pathlib
import uuid
from abc import ABC
from copy import deepcopy
from datetime import datetime, timedelta
from typing import List, Tuple, Dict, Union, Optional, Any, MutableMapping, Pattern, Match, Callable

from glQiwiApi.core import RequestManager
from glQiwiApi.core.core_mixins import ContextInstanceMixin, ToolsMixin
from glQiwiApi.core.web_hooks.dispatcher import Dispatcher
from glQiwiApi.qiwi.settings import QiwiRouter, QiwiKassaRouter
from glQiwiApi.types import (
    QiwiAccountInfo,
    Transaction,
    Statistic,
    Limit,
    Account,
    Balance,
    Identification,
    Sum,
    Card,
    Restriction,
    Commission,
    CrossRate,
    PaymentMethod,
    FreePaymentDetailsFields,
    PaymentInfo,
    OrderDetails,
    Bill,
    OptionalSum,
    P2PKeys,
    RefundBill,
    WebHookConfig
)
from glQiwiApi.types.basics import DEFAULT_CACHE_TIME
from glQiwiApi.utils import basics as api_helper
from glQiwiApi.utils.exceptions import (
    InvalidCardNumber,
    RequestError,
    InvalidData
)


def _is_copy_signal(kwargs: dict):
    try:
        return kwargs.pop("__copy_signal__")
    except KeyError:
        return False


class BaseWrapper(ABC):
    """ Base wrapper class"""
    set_current: Callable

    def __init__(self, api_access_token: Optional[str] = None,
                 phone_number: Optional[str] = None,
                 secret_p2p: Optional[str] = None,
                 without_context: bool = False,
                 cache_time: Union[float, int] = DEFAULT_CACHE_TIME,
                 validate_params: bool = False,
                 proxy: Any = None) -> None:
        """
        :param api_access_token: токен, полученный с https://qiwi.com/api
        :param phone_number: номер вашего телефона с +
        :param secret_p2p: секретный ключ, полученный с https://p2p.qiwi.com/
        :param without_context: bool, указывает, будет ли объект класса
         "глобальной" переменной или будет использована в async with контексте
        :param cache_time: Время кэширование запросов в секундах,
         по умолчанию 0, соответственно,
         запрос не будет использовать кэш по дефолту, максимальное время
         кэширование 60 секунд
        :param proxy: Прокси, которое будет использовано при создании сессии, может замедлить
            работу АПИ
        """
        if validate_params:
            self._validate_params(
                api_access_token=api_access_token,
                cache_time=cache_time,
                secret_p2p=secret_p2p,
                phone_number=phone_number,
                without_context=without_context
            )

        if isinstance(phone_number, str):
            self.phone_number = phone_number.replace('+', '')
            if self.phone_number.startswith('8'):
                self.phone_number = '7' + self.phone_number[1:]

        self._router: QiwiRouter = QiwiRouter()
        self._p2p_router: QiwiKassaRouter = QiwiKassaRouter()
        self._requests: RequestManager = RequestManager(
            without_context=without_context,
            messages=self._router.config.ERROR_CODE_NUMBERS,
            cache_time=cache_time,
            proxy=proxy
        )
        self.api_access_token = api_access_token
        self.secret_p2p = secret_p2p

        self.dispatcher = Dispatcher(loop=api_helper.take_event_loop())

        # Method from ContextInstanceMixin
        self.set_current(self)  # pragma: no cover

    def _auth_token(
            self,
            headers: MutableMapping,
            p2p: bool = False
    ) -> MutableMapping:
        """
        Make auth for API

        :param headers: dictionary
        :param p2p: boolean
        """
        headers['Authorization'] = headers['Authorization'].format(
            token=self.api_access_token if not p2p else self.secret_p2p
        )
        return headers

    @property
    def request_manager(self) -> RequestManager:
        """ Return :class:`RequestManager` """
        return self._requests

    @request_manager.setter
    def request_manager(self, manager: RequestManager):
        if not isinstance(manager, RequestManager):
            raise TypeError("Expected `RequestManager` hair, got %s" % type(manager))
        self._requests = manager

    @property
    def stripped_number(self) -> str:
        """returns number, in which the `+` sign is removed"""
        try:
            return self.phone_number.replace("+", "")
        except AttributeError:
            raise InvalidData(
                "You should pass on phone number to execute this method"
            ) from None

    @staticmethod
    def _validate_params(api_access_token: Optional[str],
                         phone_number: Optional[str],
                         secret_p2p: Optional[str],
                         without_context: bool,
                         cache_time: Union[float, int]):
        """
        Validating all parameters by `isinstance` function or `regex`

        :param api_access_token:
        :param phone_number:
        :param without_context:
        :param cache_time:
        """
        import re

        if not isinstance(api_access_token, str):
            raise InvalidData("Invalid type of api_access_token parameter, required `string`,"
                              "you have passed %s" % type(api_access_token))
        if not isinstance(secret_p2p, str):
            raise InvalidData("Invalid type of secret_p2p parameter, required `string`,"
                              "you have passed %s" % type(secret_p2p))
        if not isinstance(without_context, bool):
            raise InvalidData("Invalid type of without_context parameter, required `bool`,"
                              "you have passed %s" % type(without_context))
        if not isinstance(cache_time, (float, int)):
            raise InvalidData("Invalid type of cache_time parameter, required `bool`,"
                              "you have passed %s" % type(cache_time))

        phone_number_pattern: Pattern[str] = re.compile(
            r"^[+]?[(]?[0-9]{3}[)]?[-\s.]?[0-9]{3}[-\s.]?[0-9]{4,6}$"
        )
        match: Optional[Match[Any]] = re.fullmatch(phone_number_pattern, phone_number)

        if not match:
            raise InvalidData("Failed to verify parameter `phone_number` by regex. "
                              "Please, enter the correct phone number.")

    def __new__(cls, api_access_token: Optional[str] = None,
                phone_number: Optional[str] = None,
                secret_p2p: Optional[str] = None,
                without_context: bool = False,
                cache_time: Union[float, int] = DEFAULT_CACHE_TIME,
                *args, **kwargs):
        if not isinstance(api_access_token, str) and not isinstance(secret_p2p, str):
            if not _is_copy_signal(kwargs):
                raise RuntimeError("Cannot initialize an instance without any tokens")

        return super().__new__(cls)


class QiwiWrapper(BaseWrapper, ToolsMixin, ContextInstanceMixin["QiwiWrapper"]):
    """
    Delegates the work of QIWI API, webhooks, polling.
    Fast and versatile wrapper.

    """
    __slots__ = (
        'api_access_token',
        'phone_number',
        'secret_p2p',
        '_requests',
        '_router',
        '_p2p_router',
        'dispatcher'
    )

    async def _register_webhook(
            self,
            web_url: Optional[str],
            txn_type: int = 2
    ) -> WebHookConfig:
        """
        This method register a new webhook

        :param web_url: service url
        :param txn_type:  0 => incoming, 1 => outgoing, 2 => all
        :return: Active Hooks
        """
        url = self._router.build_url("REG_WEBHOOK")
        async for response in self._requests.fast().fetch(
                url=url,
                method='PUT',
                headers=self._auth_token(deepcopy(
                    self._router.config.DEFAULT_QIWI_HEADERS
                )),
                params={
                    'hookType': 1,
                    'param': web_url,
                    'txnType': txn_type
                },
                get_json=True
        ):
            return WebHookConfig.parse_obj(response.response_data)

    async def get_current_webhook(self) -> WebHookConfig:
        """
        Список действующих (активных) обработчиков уведомлений,
         связанных с вашим кошельком, можно получить данным запросом.
        Так как сейчас используется только один тип хука - webhook,
         то в ответе содержится только один объект данных

        """
        url = self._router.build_url("GET_CURRENT_WEBHOOK")
        async for response in self._requests.fast().fetch(
                url=url,
                method='GET',
                headers=self._auth_token(deepcopy(
                    self._router.config.DEFAULT_QIWI_HEADERS
                )),
                get_json=True
        ):
            return WebHookConfig.parse_obj(response.response_data)

    async def _send_test_notification(self) -> Dict[str, str]:
        """
        Для проверки вашего обработчика webhooks используйте данный запрос.
        Тестовое уведомление отправляется на адрес, указанный при вызове
        register_webhook

        """
        url = self._router.build_url("SEND_TEST_NOTIFICATION")
        async for response in self._requests.fast().fetch(
                url=url,
                method='GET',
                headers=self._auth_token(deepcopy(
                    self._router.config.DEFAULT_QIWI_HEADERS
                )),
                get_json=True
        ):
            return response.response_data

    async def get_webhook_secret_key(self, hook_id: str) -> str:
        """
        Каждое уведомление содержит цифровую подпись сообщения,
         зашифрованную ключом.
        Для получения ключа проверки подписи используйте данный запрос.

        :param hook_id: UUID of webhook
        :return: Base64-закодированный ключ
        """
        url = self._router.build_url(
            "GET_WEBHOOK_SECRET",
            hook_id=hook_id
        )
        async for response in self._requests.fast().fetch(
                url=url,
                method='GET',
                headers=self._auth_token(deepcopy(
                    self._router.config.DEFAULT_QIWI_HEADERS
                )),
                get_json=True
        ):
            return response.response_data.get('key')

    async def delete_current_webhook(self) -> Optional[Dict[str, str]]:
        """
        Method to delete webhook

        :return: Описание результата операции
        """
        try:
            hook = await self.get_current_webhook()
        except RequestError as ex:
            raise RequestError(
                message=" You didn't register any webhook to delete ",
                status_code='422',
                json_info=ex.json()
            ) from None

        url = self._router.build_url(
            "DELETE_CURRENT_WEBHOOK",
            hook_id=hook.hook_id
        )
        async for response in self._requests.fast().fetch(
                url=url,
                headers=self._auth_token(deepcopy(
                    self._router.config.DEFAULT_QIWI_HEADERS
                )),
                method='DELETE',
                get_json=True
        ):
            return response.response_data

    async def change_webhook_secret(self, hook_id: str) -> str:
        """
        Для смены ключа шифрования уведомлений используйте данный запрос.

        :param hook_id: UUID of webhook
        :return: Base64-закодированный ключ
        """
        url = self._router.build_url(
            "CHANGE_WEBHOOK_SECRET",
            hook_id=hook_id
        )
        async for response in self._requests.fast().fetch(
                url=url,
                headers=self._auth_token(deepcopy(
                    self._router.config.DEFAULT_QIWI_HEADERS
                )),
                method='POST',
                get_json=True
        ):
            return response.response_data.get('key')

    async def bind_webhook(
            self,
            url: Optional[str] = None,
            transactions_type: int = 2,
            *,
            send_test_notification: bool = False,
            delete_old: bool = False
    ) -> Tuple[Optional[WebHookConfig], str]:
        """
        [NON-API] EXCLUSIVE method to register new webhook or get old

        :param url: service url
        :param transactions_type: 0 => incoming, 1 => outgoing, 2 => all
        :param send_test_notification:  test_qiwi will send
         you test webhook update
        :param delete_old: boolean, if True - delete old webhook

        :return: Tuple of Hook and Base64-encoded key
        """
        key: Optional[str] = None

        if delete_old:
            await self.delete_current_webhook()

        try:
            # Try to register new webhook
            webhook = await self._register_webhook(
                web_url=url,
                txn_type=transactions_type
            )
        except (RequestError, TypeError):
            # Catching exception, if webhook already was registered
            try:
                webhook = await self.get_current_webhook()
            except RequestError as ex:
                raise RequestError(
                    message="You didn't pass on url to register new hook "
                            "and you didn't have registered webhooks",
                    status_code="422",
                    json_info=ex.json()
                ) from None
            key = await self.get_webhook_secret_key(webhook.hook_id)
            return webhook, key

        if send_test_notification:
            await self._send_test_notification()

        if not isinstance(key, str):
            key = await self.get_webhook_secret_key(webhook.hook_id)

        return webhook, key

    @api_helper.override_error_messages(
        {
            404: {
                "message": "Введена неправильный номер карты, возможно, "
                           "карта на которую вы переводите заблокирована"
            }
        }
    )
    async def _detect_mobile_number(self, phone_number: str):
        """
        Метод для получения идентификатора телефона

        https://developer.qiwi.com/ru/qiwi-wallet-personal/?python#cards
        """
        headers = deepcopy(self._router.config.DEFAULT_QIWI_HEADERS)
        headers.update(
            {
                'Content-Type': 'application/x-www-form-urlencoded'
            }
        )
        async for response in self._requests.fetch(
                url='https://qiwi.com/mobile/detect.action',
                headers=headers,
                method='POST',
                data={
                    'phone': phone_number
                },
                get_json=True
        ):
            return response.response_data.get('message')

    async def get_balance(self) -> Sum:
        """Метод для получения баланса киви"""
        if not isinstance(self.phone_number, str):
            raise InvalidData(
                "Для вызова этого метода вы должны передать номер кошелька"
            )

        headers = self._auth_token(deepcopy(
            self._router.config.DEFAULT_QIWI_HEADERS
        ))
        url = self._router.build_url(
            "GET_BALANCE",
            phone_number=self.phone_number
        )
        async for response in self._requests.fast().fetch(
                url=url,
                headers=headers,
                method='GET',
                get_json=True
        ):
            return Sum.parse_obj(
                response.response_data['accounts'][0]['balance']
            )

    async def transactions(
            self,
            rows_num: int = 50,
            operation: str = 'ALL',
            start_date: Optional[datetime] = None,
            end_date: Optional[datetime] = None
    ) -> List[Transaction]:
        """
        Метод для получения транзакций на счёту
        Более подробная документация:
        https://developer.qiwi.com/ru/qiwi-wallet-personal/?http#payments_list

        Возможные значения параметра operation:
         - 'ALL'
         - 'IN'
         - 'OUT'
         - 'QIWI_CARD'

        :param rows_num: кол-во транзакций, которые вы хотите получить
        :param operation: Тип операций в отчете, для отбора.
        :param start_date:Начальная дата поиска платежей.
                            Используется только вместе с end_date.
        :param end_date: конечная дата поиска платежей.
                            Используется только вместе со start_date.
        """
        if rows_num > 50 or rows_num <= 0:
            raise InvalidData('Можно проверять не более 50 транзакций')

        headers = self._auth_token(deepcopy(
            self._router.config.DEFAULT_QIWI_HEADERS
        ))

        payload_data = api_helper.check_dates(
            start_date=start_date,
            end_date=end_date,
            payload_data={
                'rows': rows_num,
                'operation': operation
            }
        )
        url = self._router.build_url(
            "TRANSACTIONS",
            stripped_number=self.stripped_number
        )

        async for response in self._requests.fast().fetch(
                url=url,
                params=payload_data,
                headers=headers,
                method='GET',
                get_json=True
        ):
            return api_helper.multiply_objects_parse(
                lst_of_objects=(response.response_data.get('data'),),
                model=Transaction
            )

    async def transaction_info(
            self,
            transaction_id: Union[str, int],
            transaction_type: str
    ) -> Transaction:
        """
        Метод для получения полной информации о транзакции\n
        Подробная документация:
        https://developer.qiwi.com/ru/qiwi-wallet-personal/?python#txn_info

        :param transaction_id: номер транзакции
        :param transaction_type: тип транзакции, может быть только IN или OUT
        :return: Transaction object
        """
        headers = self._auth_token(deepcopy(
            self._router.config.DEFAULT_QIWI_HEADERS
        ))
        payload_data = {
            'type': transaction_type
        }
        url = self._router.build_url(
            "TRANSACTION_INFO",
            transaction_id=transaction_id
        )
        async for response in self._requests.fast().fetch(
                url=url,
                headers=headers,
                params=payload_data,
                method='GET',
                get_json=True
        ):
            return Transaction.parse_obj(response.response_data)

    async def check_restriction(self) -> List[Restriction]:
        """
        Метод для проверки ограничений на вашем киви кошельке\n
        Подробная документация:
        https://developer.qiwi.com/ru/qiwi-wallet-personal/?python#restrictions

        :return: Список, где находиться словарь с ограничениями,
         если ограничений нет - возвращает пустой список
        """
        if not isinstance(self.phone_number, str):
            raise InvalidData(
                "Для вызова этого метода вы должны передать номер кошелька"
            )
        headers = self._auth_token(deepcopy(
            self._router.config.DEFAULT_QIWI_HEADERS
        ))
        url = self._router.build_url(
            "CHECK_RESTRICTION",
            phone_number=self.phone_number
        )
        async for response in self._requests.fast().fetch(
                url=url,
                headers=headers,
                method='GET',
                get_json=True
        ):
            return api_helper.simple_multiply_parse(
                lst_of_objects=response.response_data,
                model=Restriction
            )

    @property
    async def identification(self) -> Identification:
        """
        Функция, которая позволяет
        получить данные идентификации вашего кошелька
        Более подробная документация:
        https://developer.qiwi.com/ru/qiwi-wallet-personal/?http#ident

        :return: Response object
        """
        if not isinstance(self.phone_number, str):
            raise InvalidData(
                "Для вызова этого метода вы должны передать номер кошелька"
            )

        headers = self._auth_token(deepcopy(
            self._router.config.DEFAULT_QIWI_HEADERS
        ))
        url = self._router.build_url(
            "GET_IDENTIFICATION",
            phone_number=self.phone_number
        )
        async for response in self._requests.fast().fetch(
                url=url,
                method='GET',
                headers=headers,
                get_json=True
        ):
            return Identification.parse_obj(response.response_data)

    async def check_transaction(
            self,
            amount: Union[int, float],
            transaction_type: str = 'IN',
            sender: Optional[str] = None,
            rows_num: int = 50,
            comment: Optional[str] = None
    ) -> bool:
        """
        [ NON API METHOD ]
        Метод для проверки транзакции.\n
        Данный метод использует self.transactions(rows_num=rows_num)
        для получения платежей.\n
        Для небольшой оптимизации вы можете уменьшить rows_num задав его,
        однако это не гарантирует правильный результат
        Возможные значения параметра transaction_type:
         - 'IN'
         - 'OUT'
         - 'QIWI_CARD'


        :param amount: сумма платежа
        :param transaction_type: тип платежа
        :param sender: номер получателя
        :param rows_num: кол-во платежей, которое будет проверяться
        :param comment: комментарий, по которому будет проверяться транзакция
        :return: bool, есть ли такая транзакция в истории платежей
        """
        if transaction_type not in ['IN', 'OUT', 'QIWI_CARD']:
            raise InvalidData('Вы ввели неправильный метод транзакции')

        elif rows_num > 50 or rows_num <= 0:
            raise InvalidData('Можно проверять не более 50 транзакций')

        transactions = await self.transactions(rows_num=rows_num)

        return api_helper.check_transaction(
            transactions=transactions,
            transaction_type=transaction_type,
            comment=comment,
            amount=amount,
            sender=sender
        )

    async def get_limits(self) -> Dict[str, Limit]:
        """
        Функция для получения лимитов по счёту киви кошелька\n
        Возвращает лимиты по кошельку в виде списка,
        если лимита по определенной стране нет, то не включает его в список
        Подробная документация:
        https://developer.qiwi.com/ru/qiwi-wallet-personal/?http#limits

        :return: Limit object of Limit(pydantic)
        """
        headers = self._auth_token(deepcopy(
            self._router.config.DEFAULT_QIWI_HEADERS
        ))

        payload = {}
        limit_types = self._router.config.LIMIT_TYPES
        for index, limit_type in enumerate(limit_types):
            payload['types[' + str(index) + ']'] = limit_type
        url = self._router.build_url(
            "GET_LIMITS",
            stripped_number=self.stripped_number
        )
        async for response in self._requests.fast().fetch(
                url=url,
                get_json=True,
                headers=headers,
                params=payload,
                method='GET'
        ):
            return api_helper.parse_limits(response, Limit)

    async def get_list_of_cards(self) -> List[Card]:
        """
        Данный метод позволяет вам получить список ваших карт.

        """
        headers = self._auth_token(
            deepcopy(self._router.config.DEFAULT_QIWI_HEADERS)
        )
        async for response in self._requests.fast().fetch(
                url=self._router.build_url("GET_LIST_OF_CARDS"),
                method='GET',
                headers=headers,
                get_json=True
        ):
            return api_helper.simple_multiply_parse(
                lst_of_objects=response.response_data,
                model=Card
            )

    async def authenticate(
            self,
            birth_date: str,
            first_name: str,
            last_name: str,
            patronymic: str,
            passport: str,
            oms: Optional[str] = None,
            inn: Optional[str] = None,
            snils: Optional[str] = None
    ) -> Optional[Dict[str, bool]]:
        """
        Данный запрос позволяет отправить данные
        для идентификации вашего QIWI кошелька.
        Допускается идентифицировать не более 5 кошельков на одного владельца

        Для идентификации кошелька вы обязательно должны отправить ФИО,
        серию/номер паспорта и дату рождения.\n
        Если данные прошли проверку, то в ответе будет отображен
        ваш ИНН и упрощенная идентификация кошелька будет установлена.
        В случае если данные не прошли проверку,
        кошелек остается в статусе "Минимальный".

        :param birth_date: Дата рождения в виде строки формата 1998-02-11
        :param first_name: Ваше имя
        :param last_name: Ваша фамилия
        :param patronymic: Ваше отчество
        :param passport: Серия / Номер паспорта. Пример 4400111222
        :param oms:
        :param snils:
        :param inn:
        """

        payload = {
            "birthDate": birth_date,
            "firstName": first_name,
            "inn": inn,
            "lastName": last_name,
            "middleName": patronymic,
            "oms": oms,
            "passport": passport,
            "snils": snils
        }
        headers = self._auth_token(
            deepcopy(self._router.config.DEFAULT_QIWI_HEADERS)
        )
        url = self._router.build_url(
            "AUTHENTICATE",
            stripped_number=self.stripped_number
        )
        async for response in self._requests.fast().fetch(
                url=url,
                data=self._requests.filter_dict(payload),
                headers=headers,
                get_json=True
        ):
            if response.ok and response.status_code == 200:
                return {'success': True}

    @api_helper.override_error_messages({
        422: {
            "message": "Невозможно получить чек, из-за того, что "
                       "транзакция по такому айди не была проведена, "
                       "то есть произошла ошибка при проведении транзакции"
        }
    })
    async def get_receipt(
            self,
            transaction_id: Union[str, int],
            transaction_type: str,
            dir_path: Union[str, pathlib.Path] = None,
            file_name: Optional[str] = None
    ) -> Union[bytes, int]:
        """
        Метод для получения чека в формате байтов или файлом.\n
        Возможные значения transaction_type:
         - 'IN'
         - 'OUT'
         - 'QIWI_CARD'

        :param transaction_id: str or int, id транзакции,
         можно получить при вызове методе to_wallet, to_card
        :param transaction_type: тип транзакции может быть:
         'IN', 'OUT', 'QIWI_CARD'
        :param dir_path: путь к директории, куда вы хотите сохранить чек,
         если не указан, возвращает байты
        :param file_name: Имя файла без формата. Пример: my_receipt
        :return: pdf файл в байтовом виде или номер записанных байтов
        """
        headers = self._auth_token(deepcopy(
            self._router.config.DEFAULT_QIWI_HEADERS
        ))

        data = {
            'type': transaction_type,
            'format': 'PDF'
        }
        url = self._router.build_url(
            "GET_RECEIPT", transaction_id=transaction_id
        )
        async for response in self._requests.fast().fetch(
                url=url,
                headers=headers,
                method='GET',
                params=data,
                get_bytes=True
        ):
            if not isinstance(
                    dir_path, (str, pathlib.Path)
            ) or not isinstance(file_name, str):
                return response.response_data

            return await api_helper.save_file(
                dir_path=dir_path, file_name=file_name,
                data=response.response_data
            )

    @property
    async def account_info(self) -> QiwiAccountInfo:
        """
        Метод для получения информации об аккаунте

        """
        headers = self._auth_token(deepcopy(
            self._router.config.DEFAULT_QIWI_HEADERS
        ))
        async for response in self._requests.fast().fetch(
                url=self._router.build_url("ACCOUNT_INFO"),
                headers=headers,
                method='GET',
                get_json=True
        ):
            return QiwiAccountInfo.parse_obj(response.response_data)

    async def fetch_statistics(
            self,
            start_date: Union[datetime, timedelta],
            end_date: Union[datetime, timedelta],
            operation: str = 'ALL',
            sources: Optional[List[str]] = None
    ) -> Statistic:
        """
        Данный запрос используется для получения сводной статистики
        по суммам платежей за заданный период.\n
        Более подробная документация:
        https://developer.qiwi.com/ru/qiwi-wallet-personal/?http#payments_list

        :param start_date: Начальная дата периода статистики.
         Обязательный параметр
        :param end_date: Конечная дата периода статистики.
         Обязательный параметр
        :param operation: Тип операций, учитываемых при подсчете статистики.
         Допустимые значения:
            ALL - все операции,
            IN - только пополнения,
            OUT - только платежи,
            QIWI_CARD - только платежи по картам QIWI (QVC, QVP).
            По умолчанию ALL.
        :param sources: Источники платежа, по которым вернутся данные.
            QW_RUB - рублевый счет кошелька,
            QW_USD - счет кошелька в долларах,
            QW_EUR - счет кошелька в евро,
            CARD - привязанные и непривязанные к кошельку банковские карты,
            MK - счет мобильного оператора. Если не указан,
            учитываются все источники платежа.
        """
        headers = self._auth_token(
            deepcopy(self._router.config.DEFAULT_QIWI_HEADERS)
        )
        # Raise exception if invalid value of start_date or end_date
        api_helper.check_dates_for_statistic_request(
            start_date=start_date,
            end_date=end_date
        )
        params = {
            'startDate': api_helper.datetime_to_str_in_iso(start_date),
            'endDate': api_helper.datetime_to_str_in_iso(end_date),
            'operation': operation,
        }

        if sources:
            params.update({'sources': ' '.join(sources)})

        url = self._router.build_url(
            "FETCH_STATISTICS",
            stripped_number=self.stripped_number
        )
        async for response in self._requests.fast().fetch(
                url=url,
                params=params,
                headers=headers,
                get_json=True,
                method='GET'
        ):
            return Statistic.parse_obj(response.response_data)

    async def list_of_balances(self) -> List[Account]:
        """
        Запрос выгружает текущие балансы счетов вашего QIWI Кошелька.
        Более подробная документация:
        https://developer.qiwi.com/ru/qiwi-wallet-personal/?http#balances_list

        """
        url = self._router.build_url(
            "LIST_OF_BALANCES",
            stripped_number=self.stripped_number
        )
        async for response in self._requests.fast().fetch(
                url=url,
                headers=self._auth_token(deepcopy(
                    self._router.config.DEFAULT_QIWI_HEADERS
                )),
                method='GET',
                get_json=True
        ):
            return api_helper.simple_multiply_parse(
                response.response_data.get('accounts'), Account
            )

    @api_helper.allow_response_code(201)
    async def create_new_balance(
            self, currency_alias: str
    ) -> Optional[Dict[str, bool]]:
        """
        Запрос создает новый счет и баланс в вашем QIWI Кошельке

        :param currency_alias: Псевдоним нового счета
        :return: Возвращает значение из декоратора allow_response_code
         Пример результата, если запрос был проведен успешно: {"success": True}
        """
        headers = self._auth_token(deepcopy(
            self._router.config.DEFAULT_QIWI_HEADERS
        ))
        payload = {
            'alias': currency_alias
        }
        url = self._router.build_url(
            "CREATE_NEW_BALANCE",
            stripped_number=self.stripped_number
        )
        async for response in self._requests.fast().fetch(
                url=url,
                headers=headers,
                data=payload,
                get_json=True
        ):
            return response.response_data

    async def available_balances(self) -> List[Balance]:
        """
        Запрос отображает псевдонимы счетов,
        доступных для создания в вашем QIWI Кошельке
        Сигнатура объекта ответа:
        class Balance(BaseModel):
            alias: str
            currency: int

        """
        url = self._router.build_url(
            "AVAILABLE_BALANCES",
            stripped_number=self.stripped_number
        )
        headers = self._auth_token(deepcopy(
            self._router.config.DEFAULT_QIWI_HEADERS
        ))
        async for response in self._requests.fast().fetch(
                url=url,
                headers=headers,
                method='GET',
                get_json=True
        ):
            return api_helper.simple_multiply_parse(
                lst_of_objects=response.response_data, model=Balance)

    @api_helper.allow_response_code(204)
    async def set_default_balance(self, currency_alias: str) -> Any:
        """
        Запрос устанавливает для вашего QIWI Кошелька счет,
        баланс которого будет использоваться для фондирования
        всех платежей по умолчанию.
        Счет должен содержаться в списке счетов, получить список можно вызвав
        метод list_of_balances

        :param currency_alias: Псевдоним нового счета,
         можно получить из list_of_balances
        :return: Возвращает значение из декоратора allow_response_code
         Пример результата, если запрос был проведен успешно: {"success": True}
        """
        headers = self._auth_token(deepcopy(
            self._router.config.DEFAULT_QIWI_HEADERS
        ))
        url = self._router.build_url(
            "SET_DEFAULT_BALANCE",
            stripped_number=self.stripped_number,
            currency_alias=currency_alias
        )
        async for response in self._requests.fast().fetch(
                url=url,
                headers=headers,
                method='PATCH',
                json={'defaultAccount': True},
                get_json=True
        ):
            return response

    @property
    def transaction_handler(self):
        """
        Handler manager for default QIWI transactions,
        you can pass on lambda filter, if you want,
        but it must to return a boolean

        """
        return self.dispatcher.transaction_handler_wrapper

    @property
    def bill_handler(self):
        """
        Handler manager for P2P bills,
        you can pass on lambda filter, if you want
        But it must to return a boolean

        """
        return self.dispatcher.bill_handler_wrapper

    async def to_wallet(
            self,
            to_number: str,
            trans_sum: Union[str, float, int],
            currency: str = '643',
            comment: str = '+comment+') -> Optional[str]:
        """
        Метод для перевода денег на другой кошелек\n
        Подробная документация:
        https://developer.qiwi.com/ru/qiwi-wallet-personal/?python#p2p

        :param to_number: номер получателя
        :param trans_sum: кол-во денег, которое вы хотите перевести
        :param currency: особенный код валюты
        :param comment: комментарий к платежу
        """
        data = api_helper.set_data_to_wallet(
            data=deepcopy(self._router.config.QIWI_TO_WALLET),
            to_number=to_number,
            trans_sum=trans_sum,
            currency=currency,
            comment=comment
        )
        data.headers = self._auth_token(
            headers=data.headers
        )
        async for response in self._requests.fast().fetch(
                url=self._router.build_url("TO_WALLET"),
                json=data.json,
                headers=data.headers,
                get_json=True
        ):
            return response.response_data['transaction']['id']

    async def to_card(
            self,
            trans_sum: Union[float, int],
            to_card: str
    ) -> Optional[str]:
        """
        Метод для отправки средств на карту.
        Более подробная документация:
        https://developer.qiwi.com/ru/qiwi-wallet-personal/#cards

        :param trans_sum: сумма перевода
        :param to_card: номер карты получателя
        :return:
        """
        data = api_helper.parse_card_data(
            default_data=self._router.config.QIWI_TO_CARD,
            to_card=to_card,
            trans_sum=trans_sum,
            auth_maker=self._auth_token
        )
        privat_card_id = await self._detect_card_number(card_number=to_card)
        async for response in self._requests.fast().fetch(
                url=self._router.build_url(
                    "TO_CARD", privat_card_id=privat_card_id
                ),
                headers=data.headers,
                json=data.json,
                get_json=True
        ):
            return response.response_data.get('id')

    async def _detect_card_number(self, card_number: str) -> str:
        """
        Метод для получения идентификатора карты

        https://developer.qiwi.com/ru/qiwi-wallet-personal/?python#cards
        """
        headers = deepcopy(self._router.config.DEFAULT_QIWI_HEADERS)
        headers.update(
            {
                'Content-Type': 'application/x-www-form-urlencoded'
            }
        )
        async for response in self._requests.fetch(
                url='https://qiwi.com/card/detect.action',
                headers=headers,
                method='POST',
                data={
                    'cardNumber': card_number
                },
                get_json=True
        ):
            try:
                return response.response_data.get('message')
            except KeyError:
                raise InvalidCardNumber(
                    'Invalid card number or qiwi api is not response'
                ) from None

    async def commission(
            self,
            to_account: str,
            pay_sum: Union[int, float]
    ) -> Commission:
        """
        Возвращается полная комиссия QIWI Кошелька
        за платеж в пользу указанного провайдера
        с учетом всех тарифов по заданному набору платежных реквизитов.

        :param to_account: номер карты или киви кошелька
        :param pay_sum: сумма, за которую вы хотите узнать комиссию
        :return: Commission object
        """
        payload, special_code = api_helper.parse_commission_request_payload(
            default_data=self._router.config.COMMISSION_DATA,
            auth_maker=self._auth_token,
            to_account=to_account,
            pay_sum=pay_sum
        )
        if not isinstance(special_code, str):
            special_code = await self._detect_card_number(
                card_number=to_account
            )
        url = self._router.build_url("COMMISSION", special_code=special_code)
        async for response in self._requests.fast().fetch(
                url=url,
                headers=payload.headers,
                json=payload.json,
                get_json=True
        ):
            return Commission.parse_obj(response.response_data)

    async def get_cross_rates(self) -> List[CrossRate]:
        """
        Метод возвращает текущие курсы и кросс-курсы валют КИВИ Банка.

        """
        url = self._router.build_url("GET_CROSS_RATES")
        async for response in self._requests.fast().fetch(
                url=url,
                method='GET',
                get_json=True
        ):
            return api_helper.simple_multiply_parse(
                lst_of_objects=response.response_data.get("result"),
                model=CrossRate
            )

    async def payment_by_payment_details(
            self,
            payment_sum: Sum,
            payment_method: PaymentMethod,
            fields: FreePaymentDetailsFields,
            payment_id: Optional[str] = None
    ) -> PaymentInfo:
        """
        Оплата услуг коммерческих организаций по их банковским реквизитам.

        :param payment_id: id платежа, если не передается, используется
         uuid4
        :param payment_sum: обьект Sum, в котором указывается сумма платежа
        :param payment_method: метод платежа
        :param fields: Набор реквизитов платежа
        """
        url = self._router.build_url("SPECIAL_PAYMENT")
        headers = deepcopy(self._router.config.DEFAULT_QIWI_HEADERS)
        payload = {
            "id": payment_id if isinstance(payment_id, str) else str(
                uuid.uuid4()
            ),
            "sum": payment_sum.dict(),
            "paymentMethod": payment_method.dict(),
            "fields": fields.dict()
        }
        async for response in self._requests.fast().fetch(
                url=url,
                json=payload,
                headers=headers,
                get_json=True
        ):
            return PaymentInfo.parse_obj(response.response_data)

    async def buy_qiwi_master(self) -> PaymentInfo:
        """
        Метод для покупки пакета QIWI Мастер

        Для вызова методов API вам потребуется токен API QIWI Wallet
        с разрешениями на следующие действия:

        1. Управление виртуальными картами,
        2. Запрос информации о профиле кошелька,
        3. Просмотр истории платежей,
        4. Проведение платежей без SMS.

        Эти права вы можете выбрать при создании нового апи токена,
        чтобы пользоваться апи QIWI Master
        """
        url = self._router.build_url("BUY_QIWI_MASTER")
        payload = api_helper.qiwi_master_data(self.stripped_number,
                                              self._router.config.QIWI_MASTER)
        async for response in self._requests.fast().fetch(
                url=url,
                json=payload,
                method='POST',
                headers=self._auth_token(deepcopy(
                    self._router.config.DEFAULT_QIWI_HEADERS
                )),
                get_json=True
        ):
            return PaymentInfo.parse_obj(response.response_data)

    async def __pre_qiwi_master_request(
            self,
            card_alias: str = 'qvc-cpa'
    ) -> OrderDetails:
        """
        Метод для выпуска виртуальной карты QIWI Мастер

        :param card_alias: Тип карты
        :return: OrderDetails
        """
        url = self._router.build_url(
            "PRE_QIWI_REQUEST",
            stripped_number=self.stripped_number
        )
        async for response in self._requests.fast().fetch(
                url=url,
                headers=self._auth_token(deepcopy(
                    self._router.config.DEFAULT_QIWI_HEADERS
                )),
                json={"cardAlias": card_alias},
                method='POST',
                get_json=True
        ):
            return OrderDetails.parse_obj(response.response_data)

    async def _confirm_qiwi_master_request(
            self,
            card_alias: str = 'qvc-cpa'
    ) -> OrderDetails:
        """
        Подтверждение заказа выпуска карты

        :param card_alias: Тип карты
        :return: OrderDetails
        """
        details = await self.__pre_qiwi_master_request(card_alias)
        url = self._router.build_url(
            "_CONFIRM_QIWI_MASTER",
            stripped_number=self.stripped_number,
            order_id=details.order_id
        )
        async for response in self._requests.fast().fetch(
                url=url,
                headers=self._auth_token(deepcopy(
                    self._router.config.DEFAULT_QIWI_HEADERS
                )),
                method='PUT',
                get_json=True
        ):
            return OrderDetails.parse_obj(response.response_data)

    async def __buy_new_qiwi_card(
            self,
            **kwargs
    ) -> Optional[OrderDetails]:
        """
        Покупка карты, если она платная

        :param kwargs:
        :return: OrderDetails
        """
        kwargs.update(data=self._router.config.QIWI_MASTER)
        url, payload = api_helper.new_card_data(**kwargs)
        async for response in self._requests.fast().fetch(
                url=url,
                json=payload,
                headers=self._auth_token(deepcopy(
                    self._router.config.DEFAULT_QIWI_HEADERS
                )),
                get_json=True
        ):
            return OrderDetails.parse_obj(response.response_data)

    async def issue_qiwi_master_card(
            self,
            card_alias: str = 'qvc-cpa'
    ) -> Optional[OrderDetails]:
        """
        Выпуск новой карты, используя Qiwi Master API

        При выпуске карты производиться 3, а возможно 3 запроса,
        а именно по такой схеме:
            - __pre_qiwi_master_request - данный метод создает заявку
            - _confirm_qiwi_master_request - подтверждает выпуск карты
            - __buy_new_qiwi_card - покупает новую карту,
              если такая карта не бесплатна


        Подробная документация:

        https://developer.qiwi.com/ru/qiwi-wallet-personal/#qiwi-master-issue-card

        :param card_alias: Тип карты
        :return: OrderDetails
        """
        pre_response = await self._confirm_qiwi_master_request(card_alias)
        if pre_response.status == 'COMPLETED':
            return pre_response
        return await self.__buy_new_qiwi_card(
            ph_number=self.stripped_number,
            order_id=pre_response.order_id
        )

    async def _cards_qiwi_master(self):
        """
        Метод для получение списка всех ваших карт QIWI Мастер

        """
        url = self._router.build_url("CARDS_QIWI_MASTER")
        async for response in self._requests.fast().fetch(
                url=url,
                headers=self._auth_token(deepcopy(
                    self._router.config.DEFAULT_QIWI_HEADERS
                )),
                method='GET',
                get_json=True
        ):
            return response

    async def reject_p2p_bill(self, bill_id: str) -> Bill:
        """
        Метод для отмены транзакции.

        :param bill_id: номер p2p транзакции
        :return: Bill obj
        """
        if not self.secret_p2p:
            raise InvalidData('Не задан p2p токен')
        data = deepcopy(self._p2p_router.config.P2P_DATA)
        headers = self._auth_token(data.headers, p2p=True)
        url = f'https://api.qiwi.com/partner/bill/v1/bills/{bill_id}/reject'
        async for response in self._requests.fast().fetch(
                url=url,
                method='POST',
                headers=headers,
                get_json=True
        ):
            return Bill.parse_obj(response.response_data)

    async def check_p2p_bill_status(self, bill_id: str) -> str:
        """
        Метод для проверки статуса p2p транзакции.\n
        Возможные типы транзакции: \n
        WAITING	Счет выставлен, ожидает оплаты	\n
        PAID	Счет оплачен	\n
        REJECTED	Счет отклонен\n
        EXPIRED	Время жизни счета истекло. Счет не оплачен\n
        Более подробная документация:
        https://developer.qiwi.com/ru/p2p-payments/?shell#invoice-status

        :param bill_id: номер p2p транзакции
        :return: статус транзакции строкой
        """
        if not self.secret_p2p:
            raise InvalidData('Не задан p2p токен')

        data = deepcopy(self._router.config.P2P_DATA)
        headers = self._auth_token(data.headers, p2p=True)
        url = self._p2p_router.build_url(
            "CHECK_P2P_BILL_STATUS",
            bill_id=bill_id
        )
        async for response in self._requests.fast().fetch(
                url=url,
                method='GET',
                headers=headers,
                get_json=True
        ):
            return Bill.parse_obj(response.response_data).status.value

    async def create_p2p_bill(
            self,
            amount: Union[int, float],
            bill_id: Optional[str] = None,
            comment: Optional[str] = None,
            life_time: Optional[datetime] = None,
            theme_code: Optional[str] = None,
            pay_source_filter: Optional[List[str]] = None
    ) -> Bill:
        """
        Метод для выставление p2p счёта.
        Надежный способ для интеграции.
        Параметры передаются server2server с использованием авторизации.
        Возможные значения pay_source_filter:
          - 'qw'
          - 'card'
          - 'mobile'

        :param amount: сумма платежа
        :param bill_id: уникальный номер транзакции, если не передан,
         генерируется автоматически,
        :param life_time: дата, до которой счет будет доступен для оплаты.
        :param comment: комментарий к платежу
        :param theme_code: специальный код темы
        :param pay_source_filter: При открытии формы будут отображаться
         только указанные способы перевода
        """
        from glQiwiApi.core import constants

        if not self.secret_p2p:
            raise InvalidData('Не задан p2p токен')

        if not isinstance(bill_id, (str, int)):
            bill_id = str(uuid.uuid4())

        _life_time = api_helper.datetime_to_str_in_iso(
            constants.DEFAULT_BILL_TIME if not life_time else life_time
        )

        data = deepcopy(self._p2p_router.config.P2P_DATA)

        headers = self._auth_token(data.headers, p2p=True)

        payload = api_helper.set_data_p2p_create(
            wrapped_data=data,
            amount=amount,
            comment=comment,
            theme_code=theme_code,
            pay_source_filter=pay_source_filter,
            life_time=str(_life_time)
        )
        url = self._p2p_router.build_url(
            "CREATE_P2P_BILL",
            bill_id=bill_id
        )
        async for response in self._requests.fast().fetch(
                url=url,
                json=payload,
                headers=headers,
                method='PUT',
                get_json=True
        ):
            return Bill.parse_obj(response.response_data)

    async def get_bills(self, rows_num: int) -> List[Bill]:
        """
        Метод получения списка неоплаченных счетов вашего кошелька.
        Список строится в обратном хронологическом порядке.
        По умолчанию, список разбивается на страницы по 50 элементов в каждой,
        но вы можете задать другое количество элементов (не более 50).
        В запросе можно использовать фильтры по времени выставления счета,
        начальному идентификатору счета.
        """
        headers = self._auth_token(
            deepcopy(self._router.config.DEFAULT_QIWI_HEADERS)
        )
        if rows_num > 50:
            raise InvalidData('Можно получить не более 50 счетов')

        params = {
            'rows': rows_num,
            'statuses': 'READY_FOR_PAY'
        }
        async for response in self._requests.fast().fetch(
                url=self._router.build_url("GET_BILLS"),
                headers=headers,
                method='GET',
                params=params,
                get_json=True
        ):
            return api_helper.simple_multiply_parse(
                response.response_data.get("bills"), Bill
            )

    async def refund_bill(
            self,
            bill_id: Union[str, int],
            refund_id: Union[str, int],
            json_bill_data: Union[OptionalSum, Dict[str, Union[str, int]]]
    ) -> RefundBill:
        """
        Метод позволяет сделать возврат средств по оплаченному счету.
        в JSON-теле запроса параметра json_bill_data:\n
         amount.value - сумма возврата. \n
         amount.currency - валюта возврата.
        Может быть словарем или объектом OptionalSum\n
         Пример словаря: {
        "amount": {
            "currency": "RUB",
            "value": 1
            }
        }

        :param bill_id: уникальный идентификатор счета в системе мерчанта
        :param refund_id: уникальный идентификатор возврата в системе мерчанта.
        :param json_bill_data:
        :return: RefundBill object
        """
        url = self._router.build_url(
            "REFUND_BILL",
            refund_id=refund_id,
            bill_id=bill_id
        )
        headers = self._auth_token(
            deepcopy(self._router.config.DEFAULT_QIWI_HEADERS), p2p=True
        )
        async for response in self._requests.fast().fetch(
                url=url,
                headers=headers,
                method='PUT',
                json=json_bill_data if isinstance(
                    json_bill_data,
                    dict
                ) else json_bill_data.json(),
                get_json=True
        ):
            return RefundBill.parse_obj(response.response_data)

    async def create_p2p_keys(
            self,
            key_pair_name: str,
            server_notification_url: Optional[str] = None) -> P2PKeys:
        """
        Метод создает новые p2p ключи

        :param key_pair_name: Название пары токенов P2P (произвольная строка)
        :param server_notification_url: url для вебхуков, необязательный
         параметр
        """
        url = self._router.build_url("CREATE_P2P_KEYS")
        headers = self._auth_token(
            deepcopy(self._router.config.DEFAULT_QIWI_HEADERS), p2p=True
        )
        data = {
            'keysPairName': key_pair_name,
            'serverNotificationsUrl': server_notification_url
        }
        async for response in self._requests.fast().fetch(
                url=url,
                headers=headers,
                json=data,
                get_json=True
        ):
            return P2PKeys.parse_obj(response.response_data)



from pydantic import Field

from glQiwiApi.types.base import Base


class Restriction(Base):
    code: str = Field(..., alias="restrictionCode")
    description: str = Field(..., alias="restrictionDescription")


__all__ = ('Restriction',)



from __future__ import annotations

from typing import TYPE_CHECKING

from pydantic import BaseModel

if TYPE_CHECKING:
    from glQiwiApi.qiwi.client import QiwiWrapper


class Base(BaseModel):

    @property
    def client(self) -> QiwiWrapper:
        """ Returning an instance of :class:`QiwiWrapper` """
        from glQiwiApi.qiwi.client import QiwiWrapper

        instance = QiwiWrapper.get_current()

        if instance is None:
            raise RuntimeError("Can't get client instance from context. "
                               "You can fix it with setting current instance: "
                               "'QiwiWrapper.set_current(wrapper_instance)'")
        return instance



import sys

from .qiwi import QiwiWrapper, QiwiMaps  # NOQA
from .utils.basics import sync, async_as_sync  # NOQA
from .utils.exceptions import *  # NOQA
from .yoo_money import YooMoneyAPI  # NOQA

__version__ = '1.0.0'

__all__ = (
        (
            'QiwiWrapper',
            'YooMoneyAPI',
            'QiwiMaps',
            'sync',
            'async_as_sync'
        ) + utils.exceptions.__all__  # NOQA
)


class VersionError(Exception):
    """
    Ошибка возникает, если ваша версия python не поддерживается библиотекой

    """


if not sys.version_info[:2] >= (3, 7):
    raise VersionError(
        "Ваша версия python не поддерживается библиотекой glQiwiApi."
        "Минимальная версия для использования python 3.7"
    )



import concurrent.futures as futures
from typing import Union, TypeVar, Optional, Callable, Any, MutableMapping, \
    Dict, NewType

from .basics import (
    Type,
    Sum,
    Commission,
    OptionalSum
)
from .particular import (
    Response,
    WrapperData
)
from .qiwi_types import (
    Bill,
    BillError,
    Statistic,
    Balance,
    Identification,
    Limit,
    Account,
    QiwiAccountInfo,
    Transaction,
    PaymentInfo,
    OrderDetails,
    RefundBill,
    Polygon,
    Terminal,
    Partner,
    WebHookConfig,
    WebHook,
    Notification,
    P2PKeys,
    CrossRate,
    FreePaymentDetailsFields,
    PaymentMethod,
    Card,
    Restriction
)
from .yoomoney_types import (
    OperationType,
    OperationDetails,
    PreProcessPaymentResponse,
    Operation,
    Payment,
    IncomingTransaction,
    AccountInfo
)

ALL_TYPES = Union[
    Transaction, QiwiAccountInfo, Account, Limit,
    Bill, BillError, Statistic, Balance, Identification,
    AccountInfo, Operation, OperationDetails,
    PreProcessPaymentResponse, Payment, IncomingTransaction,
    Terminal, Partner
]

Executors = TypeVar(
    'Executors',
    futures.ThreadPoolExecutor,
    futures.ProcessPoolExecutor,
    Optional[None]
)

FuncT = TypeVar('FuncT', bound=Callable[..., Any])

N = TypeVar('N')

__all__ = (
    'QiwiAccountInfo',
    'Transaction',
    'Bill',
    'BillError',
    'P2PKeys',
    'Statistic',
    'Limit',
    'Account',
    'Identification',
    'Balance',
    'AccountInfo',
    'OperationType',
    'Operation',
    'OperationDetails',
    'PreProcessPaymentResponse',
    'Payment',
    'IncomingTransaction',
    'Response',
    'WrapperData',
    'Sum',
    'Type',
    'OptionalSum',
    'Commission',
    'ALL_TYPES',
    'PaymentInfo',
    'OrderDetails',
    'RefundBill',
    'Polygon',
    'Terminal',
    'Partner',
    'WebHookConfig',
    'WebHook',
    'Notification',
    'Executors',
    'FuncT',
    'N',
    'CrossRate',
    'FreePaymentDetailsFields',
    'PaymentMethod',
    'Card',
    'Restriction'
)



from __future__ import annotations

import functools
import time
from dataclasses import dataclass
from functools import lru_cache
from typing import Optional, Any, Union, Dict, List

from glQiwiApi.core.abstracts import AbstractRouter
from glQiwiApi.types import WrapperData
from glQiwiApi.utils.basics import check_api_method

__all__ = ('get_settings', "QiwiRouter", "QiwiKassaRouter")


@lru_cache()
def get_settings() -> QiwiSettings:
    settings = QiwiSettings()
    return settings


class QiwiRouter(AbstractRouter):
    """Class, that delegates building the right paths, except p2p"""
    __head__ = "https://edge.qiwi.com"

    @functools.lru_cache()
    def build_url(self, api_method: str, **kwargs: Any) -> str:
        check_api_method(api_method)
        tail: str = getattr(self.config, api_method, None)
        pre_build_url = self.__head__ + tail
        return super()._format_url_kwargs(pre_build_url, **kwargs)

    def setup_config(self) -> Any:
        return get_settings()


class QiwiKassaRouter(QiwiRouter):
    """ QIWI P2P router"""
    __head__ = "https://api.qiwi.com/partner/bill/v1/bills"


@dataclass
class QiwiSettings:
    # Qiwi Wallet
    TO_CARD: str = '/sinap/api/v2/terms/{privat_card_id}'

    GET_BALANCE: str = '/funding-sources/v2/persons/{phone_number}/accounts'

    TRANSACTIONS: str = '/payment-history/v2/persons/{stripped_number}/payments'

    TO_WALLET: str = '/sinap/api/v2/terms/99/payments'

    TRANSACTION_INFO: str = '/payment-history/v1/transactions/{transaction_id}'

    CHECK_RESTRICTION: str = '/person-profile/v1/persons/{phone_number}/status/restrictions'

    GET_IDENTIFICATION: str = '/identification/v1/persons/{phone_number}/identification'

    GET_LIMITS: str = '/qw-limits/v1/persons/{stripped_number}/actual-limits'

    GET_LIST_OF_CARDS: str = '/cards/v1/cards'

    AUTHENTICATE: str = '/identification/v1/persons/{stripped_number}/identification'

    GET_BILLS: str = '/checkout-api/api/bill/search'

    GET_RECEIPT: str = '/payment-history/v1/transactions/{transaction_id}/cheque/file'

    COMMISSION: str = '/sinap/providers/{special_code}/onlineCommission'

    ACCOUNT_INFO: str = '/person-profile/v1/profile/current'

    FETCH_STATISTICS: str = '/payment-history/v2/persons/{stripped_number}/payments/total'

    LIST_OF_BALANCES: str = '/funding-sources/v2/persons/{stripped_number}/accounts'

    CREATE_NEW_BALANCE: str = '/funding-sources/v2/persons/{stripped_number}/accounts'

    AVAILABLE_BALANCES: str = '/funding-sources/v2/persons/{stripped_number}/accounts/offer'

    SET_DEFAULT_BALANCE: str = '/funding-sources/v2/persons/{stripped_number}/accounts/{currency_alias}'

    BUY_QIWI_MASTER: str = '/sinap/api/v2/terms/28004/payments'

    _CONFIRM_QIWI_MASTER: str = '/cards/v2/persons/{stripped_number}/orders/{order_id}/submit'

    CARDS_QIWI_MASTER: str = '/cards/v1/cards/?vas-alias=qvc-master'

    PRE_QIWI_REQUEST: str = '/cards/v2/persons/{number}/orders'

    GET_CROSS_RATES: str = '/sinap/crossRates'

    SPECIAL_PAYMENT: str = '/sinap/api/v2/terms/1717/payments'

    # Qiwi P2P
    CREATE_P2P_BILL: str = '/{bill_id}'
    CHECK_P2P_BILL_STATUS: str = '/{bill_id}'
    REJECT_P2P_BILL: str = '/{bill_id}/reject'
    REFUND_BILL: str = '/{bill_id}/refunds/{refund_id}'
    CREATE_P2P_KEYS: str = '/widgets-api/api/p2p/protected/keys/create'

    # Webhooks
    REG_WEBHOOK: str = '/payment-notifier/v1/hooks'

    GET_CURRENT_WEBHOOK: str = '/payment-notifier/v1/hooks/active'

    SEND_TEST_NOTIFICATION: str = '/payment-notifier/v1/hooks/test'

    GET_WEBHOOK_SECRET: str = '/payment-notifier/v1/hooks/{hook_id}/key'

    DELETE_CURRENT_WEBHOOK: str = '/payment-notifier/v1/hooks/{hook_id}'

    CHANGE_WEBHOOK_SECRET: str = '/payment-notifier/v1/hooks/{hook_id}/newkey'

    # Payload data
    DEFAULT_QIWI_HEADERS: Optional[Dict[str, Union[str, int]]] = None

    ERROR_CODE_NUMBERS: Optional[Dict[str, Union[str, int]]] = None

    P2P_QIWI_HEADERS: Optional[Dict[str, Union[str, int]]] = None

    def __init__(self, *args, **kwargs):
        self.DEFAULT_QIWI_HEADERS = {
            'Content-Type': 'application/json',
            'Accept': 'application/json',
            'Authorization': 'Bearer {token}',
            'Host': 'edge.test_qiwi.com',
        }
        self.P2P_QIWI_HEADERS = {
            'Authorization': 'Bearer {token}',
            'Accept': 'application/json',
            'Content-Type': 'application/json',
        }
        self.ERROR_CODE_NUMBERS = {
            "400": "Недостаточно средств для проведения операции",
            "401": "Неверный токен или истек срок действия токена API",
            "403": "Нет прав на данный запрос"
                   "(недостаточно разрешений у токена API)",
            "404": "Не найдена транзакция или "
                   "отсутствуют платежи с указанными признаками",
            "423": "Слишком много запросов, сервис временно недоступен",
            "422": "Неправильно указаны домен/подсеть/хост"
                   " веб-хука(в параметре new_url для URL веб-хука), "
                   "неправильно указаны тип хука или тип транзакции, "
                   "попытка создать хук при наличии уже созданного",
            "405": "Ошибка, связанная с типом запроса к апи,"
                   "обратитесь к разработчику или откройте issue",
            "500": "Внутренняя ошибка сервиса",
            "400_special_bad_proxy": "Ошибка, связанная с использованием прокси"
        }
        self.QIWI_MASTER = {
            "id": str(int(time.time() * 1000)),
            "sum": {
                "amount": 2999,
                "currency": "643"
            },
            "paymentMethod": {
                "type": "Account",
                "accountId": "643"
            },
            "comment": "test",
            "fields": {
                "account": "",
                "vas_alias": "qvc-master"
            }
        }

        for key, value in kwargs.items():
            setattr(self, key, value)
        for arg in args:
            setattr(self, arg, arg)

        self.LIMIT_TYPES = [
            'TURNOVER',
            'REFILL',
            'PAYMENTS_P2P',
            'PAYMENTS_PROVIDER_INTERNATIONALS',
            'PAYMENTS_PROVIDER_PAYOUT',
            'WITHDRAW_CASH'
        ]
        self.QIWI_TO_CARD.headers = self.DEFAULT_QIWI_HEADERS
        self.P2P_DATA.headers = self.DEFAULT_QIWI_HEADERS
        self.QIWI_TO_WALLET.headers = self.DEFAULT_QIWI_HEADERS
        self.COMMISSION_DATA.headers = self.DEFAULT_QIWI_HEADERS

    QIWI_TO_CARD: WrapperData = WrapperData(
        json={
            "id": str(int(time.time() * 1000)),
            "sum": {
                "amount": "", "currency": "643"
            },
            "paymentMethod": {
                "type": "Account", "accountId": "643"
            },
            "fields": {
                "account": ""
            }
        },
        headers=DEFAULT_QIWI_HEADERS
    )

    P2P_DATA: WrapperData = WrapperData(
        json={
            "amount": {
                "currency": "RUB",
                "value": "{amount}"
            },
            "expirationDateTime": "",
            "comment": "{comment}",
            "customFields": {
                "paySourcesFilter": "qw",
                "themeCode": "Yvan-YKaSh",
            }
        },

        headers=P2P_QIWI_HEADERS
    )

    QIWI_TO_WALLET: WrapperData = WrapperData(
        json={
            "id": str(int(time.time() * 1000)),
            "sum": {
                "amount": "",
                "currency": ""
            },
            "paymentMethod": {
                "type": "Account",
                "accountId": "643"
            },
            "comment": "",
            "fields": {
                "account": ""
            }
        },
        headers=DEFAULT_QIWI_HEADERS
    )

    LIMIT_TYPES: Optional[List[str]] = None

    COMMISSION_DATA: WrapperData = WrapperData(
        json={
            "account": "",
            "paymentMethod": {
                "type": "Account",
                "accountId": "643"
            },
            "purchaseTotals":
                {
                    "total":
                        {"amount": "",
                         "currency": "643"
                         }
                }},
        headers=DEFAULT_QIWI_HEADERS
    )

    QIWI_MASTER: Optional[Dict[str, Union[int, float, str]]] = None



from datetime import datetime
from typing import Optional, Union

from pydantic import Field, Extra

from glQiwiApi.types.base import Base
from glQiwiApi.types.basics import OptionalSum


class Customer(Base):
    """ Object: Customer """
    phone: Optional[str] = None
    email: Optional[str] = None
    account: Optional[str] = None


class BillStatus(Base):
    """ Object: BillStatus """
    value: str
    changed_datetime: datetime = Field(alias="changedDateTime")


class CustomFields(Base):
    """ Object: CustomFields """
    pay_sources_filter: Optional[str] = Field(alias="paySourcesFilter",
                                              default=None)
    theme_code: Optional[str] = Field(alias="themeCode", default=None)


class BillError(Base):
    """ Object: BillError """
    service_name: str = Field(alias="serviceName")
    error_code: str = Field(alias="errorCode")
    description: str
    user_message: str = Field(alias="userMessage")
    datetime: str = Field(alias="dateTime")
    trace_id: str = Field(alias="traceId")


class Bill(Base):
    """ Object: Bill """
    site_id: str = Field(alias="siteId")
    bill_id: str = Field(alias="billId")
    amount: OptionalSum
    status: BillStatus
    creation_date_time: datetime = Field(alias="creationDateTime")
    expiration_date_time: datetime = Field(alias="expirationDateTime")
    pay_url: Optional[str] = Field(alias="payUrl", default=None)
    custom_fields: Optional[
        CustomFields
    ] = Field(alias="customFields", default=None)
    customer: Optional[Customer] = None

    class Config:
        extra = Extra.allow

        def __str__(self) -> str:
            return "<Config pydantic model {Bill}>"

        def __repr__(self) -> str:
            return self.__str__()

    @property
    async def paid(self) -> bool:
        """
        Checking p2p payment

        """
        return (await self.client.check_p2p_bill_status(
            bill_id=self.bill_id
        )) == 'PAID'


class RefundBill(Base):
    """
    Модель счёта киви апи

    """
    amount: OptionalSum
    datetime: datetime
    refund_id: str = Field(..., alias="refundId")
    status: str

    def as_str(self) -> str:
        return f"№{self.refund_id} {self.status} {self.amount} {self.datetime}"

    def get_value(self) -> Union[float, int]:
        return self.amount.value


class Notification(Base):
    """Object: Notification"""

    version: str = Field(..., alias="version")
    bill: Bill = Field(..., alias="bill")

    def __str__(self) -> str:
        return f"#{self.bill.bill_id} {self.bill.amount} {self.bill.status} "

    def __repr__(self) -> str:
        return self.__str__()


class P2PKeys(Base):
    public_key: str = Field(..., alias="PublicKey")
    secret_key: str = Field(..., alias="SecretKey")


__all__ = (
    'Bill',
    'BillError',
    'RefundBill',
    'Notification',
    'P2PKeys'
)



"""
Managing polling and webhooks
"""
from __future__ import annotations

import asyncio
import inspect
import logging
import types
from datetime import datetime, timedelta
from typing import Union, Optional, List, \
    Callable, Awaitable, Dict, TypeVar, TYPE_CHECKING, Coroutine, Any

from aiohttp import ClientTimeout, web

from glQiwiApi.core.builtin import BaseProxy, logger
from glQiwiApi.core.constants import DEFAULT_TIMEOUT
from glQiwiApi.core.web_hooks import server
from glQiwiApi.core.web_hooks.config import Path
from glQiwiApi.core.web_hooks.dispatcher import Dispatcher
from glQiwiApi.types import Transaction
from glQiwiApi.utils import basics as api_helper
from glQiwiApi.utils.exceptions import NoUpdatesToExecute

__all__ = ['start_webhook', 'start_polling']

if TYPE_CHECKING:
    from glQiwiApi.qiwi.client import QiwiWrapper

T = TypeVar("T")


def start_webhook(client: QiwiWrapper, *, host: str = "localhost",
                  port: int = 8080,
                  path: Optional[Path] = None,
                  on_startup: Optional[
                      Callable[
                          [QiwiWrapper], Awaitable[None]
                      ]] = None,
                  on_shutdown: Optional[
                      Callable[
                          [QiwiWrapper], Awaitable[None]
                      ]] = None,
                  tg_app: Optional[BaseProxy] = None,
                  app: Optional["web.Application"] = None):
    """
    Blocking function that listens for webhooks

    :param client:
    :param host: server host
    :param port: server port that open for tcp/ip trans.
    :param path: path for qiwi that will send requests
    :param app: pass web.Application
    :param on_startup: coroutine,which will be executed on startup
    :param on_shutdown: coroutine, which will be executed on shutdown
    :param tg_app: builtin TelegramWebhookProxy or other
          or class, that inherits from BaseProxy and deal with aiogram updates
    :param logger_config:
    """
    executor = Executor(client, tg_app=tg_app)
    _setup_callbacks(executor, on_startup, on_shutdown)
    executor.start_webhook(host=host, port=port, path=path, app=app)


def start_polling(client: QiwiWrapper, *, get_updates_from: datetime = datetime.now(),
                  timeout: Union[float, int, ClientTimeout] = 5,
                  on_startup: Optional[
                      Callable[
                          [QiwiWrapper], Any
                      ]] = None,
                  on_shutdown: Optional[
                      Callable[
                          [QiwiWrapper], Any
                      ]] = None,
                  tg_app: Optional[BaseProxy] = None,
                  **kwargs):
    """
    Setup for long-polling mode

    :param client:
    :param get_updates_from: date from which will be polling,
         if it's None, polling will skip all updates
    :param timeout: timeout of polling in seconds, if the timeout is too small,
         the API can throw an exception
    :param kwargs: keyword arguments that you can pass
         into the aiogram `start_polling` method
    :param on_startup: function or coroutine,
         which will be executed on startup
    :param on_shutdown: function or coroutine,
         which will be executed on shutdown
    :param tg_app: builtin TelegramPollingProxy or other
         class, that inherits from BaseProxy, deal with aiogram updates
    """
    executor = Executor(client, tg_app=tg_app)
    _setup_callbacks(executor, on_startup, on_shutdown)
    executor.start_polling(
        get_updates_from=get_updates_from,
        timeout=timeout,
        **kwargs
    )


async def _inspect_and_execute_callback(client, callback: Callable):
    if inspect.iscoroutinefunction(callback):
        await callback(client)
    else:
        callback(client)


def _setup_callbacks(
        executor: Executor,
        on_startup: Optional[Callable] = None,
        on_shutdown: Optional[Callable] = None
):
    """
    Function, which setup callbacks and set it to dispatcher object

    :param executor:
    :param on_startup:
    :param on_shutdown:
    """
    if on_startup is not None:
        executor["on_startup"] = on_startup
    if on_shutdown is not None:
        executor["on_shutdown"] = on_shutdown


def parse_timeout(
        timeout: Union[float, int, ClientTimeout]
) -> float:
    """
    Parse timeout

    :param timeout:
    """
    if isinstance(timeout, float):
        return timeout
    elif isinstance(timeout, int):
        return float(timeout)
    elif isinstance(timeout, ClientTimeout):
        return timeout.total or DEFAULT_TIMEOUT.total
    else:
        raise ValueError("Timeout must be float, int or ClientTimeout. You have "
                         f"passed on {type(timeout)}")


class Executor:
    """
    Provides normal work of webhooks and polling

    """

    def __init__(self, client: QiwiWrapper, tg_app: Optional[BaseProxy]):
        """

        :param client: instance of BaseWrapper
        :param tg_app: optional proxy to connect aiogram polling/webhook mode
        """
        self.dispatcher: Dispatcher = client.dispatcher
        self._loop: asyncio.AbstractEventLoop = client.dispatcher._loop
        self._logger_config: Dict[str, Union[List[logging.Handler], int]] = {
            "handlers": [logger.InterceptHandler()],
            "level": logging.DEBUG
        }
        self.tg_app: Optional[BaseProxy] = tg_app
        self._polling: bool = False
        self.offset: Optional[int] = None
        self.offset_start_date: Optional[datetime] = None
        self.offset_end_date: Optional[datetime] = None
        self.client: QiwiWrapper = client
        self._on_startup_calls: List[Callable] = []
        self._on_shutdown_calls: List[Callable] = []

        from glQiwiApi import QiwiWrapper
        QiwiWrapper.set_current(client)

    def __setitem__(self, key: str, callback: Callable):
        if key not in ["on_shutdown", "on_startup"]:
            raise RuntimeError()

        if not isinstance(callback, types.FunctionType):
            raise RuntimeError("Invalid type of callback")

        if key == "on_shutdown":
            self._on_shutdown_calls.append(callback)
        else:
            self._on_startup_calls.append(callback)

    async def _pre_process(self, get_updates_from: Optional[datetime]):
        """
        Preprocess method, which set start date and end date of polling
        :param get_updates_from: date from which will be polling
        """
        try:
            current_time = datetime.now()
            assert isinstance(get_updates_from, datetime)
            assert (
                           current_time - get_updates_from
                   ).total_seconds() > 0
        except AssertionError as ex:
            raise ValueError(
                "Invalid value of get_updates_from, it must "
                "be instance of datetime and no more than  the current time"
            ) from ex

        self.offset_end_date = current_time

        if self.offset_start_date is None:
            self.offset_start_date = get_updates_from
        else:
            self.offset_start_date = current_time - timedelta(milliseconds=1)

    async def _get_history(self) -> List[Transaction]:
        """
        Get history by call 'transactions' method from QiwiWrapper.
        If history is empty or not all transactions not isinstance
         class Transaction - raise exception

        """
        history = await self.client.transactions(
            end_date=self.offset_end_date,
            start_date=self.offset_start_date
        )

        if not history or not all(
                isinstance(txn, Transaction) for txn in history):
            raise NoUpdatesToExecute()

        return history

    async def _pool_process(
            self,
            get_updates_from: Optional[datetime]
    ):
        """
        Method, which manage pool process

        :param get_updates_from: date from which will be polling
        """
        await self._pre_process(get_updates_from)
        try:
            history: List[Transaction] = await self._get_history()
        except NoUpdatesToExecute:
            return

        last_payment: Transaction = history[0]
        last_txn_id: int = last_payment.transaction_id

        if self.offset is None:
            first_payment: Transaction = history[-1]
            self.offset = first_payment.transaction_id - 1

        await self._parse_history_and_process_events(
            history=history,
            last_payment_id=last_txn_id
        )

    async def _start_polling(self, **kwargs):
        """
        Blocking method, which start polling process

        :param kwargs:
        """
        self._polling = True
        timeout: float = parse_timeout(kwargs.pop("timeout"))
        while self._polling:
            try:
                await self._pool_process(**kwargs)
            except Exception as ex:
                self.dispatcher.logger.error("Handle `%s`. Set a smaller timeout or open issue. "
                                             "Sleeping %s seconds", repr(ex), timeout + 100)
                timeout += 100
            await asyncio.sleep(timeout)

    def _on_shutdown(self):
        """
        On shutdown, we gracefully cancel all tasks, close event loop
        and call `close` method to clear resources
        """
        coroutines: List[Coroutine] = [self.goodbye(), self.client.close()]
        if isinstance(self.tg_app, BaseProxy):
            coroutines.append(self._shutdown_tg_app())
        self._loop.run_until_complete(
            asyncio.gather(*coroutines, loop=self._loop)
        )

    async def _shutdown_tg_app(self):
        """
        Gracefully shutdown tg application

        """
        self.tg_app.dispatcher.stop_polling()
        await self.tg_app.dispatcher.storage.close()
        await self.tg_app.dispatcher.storage.wait_closed()
        await self.tg_app.dispatcher.bot.session.close()

    async def _parse_history_and_process_events(
            self,
            history: List[Transaction],
            last_payment_id: int
    ):
        """
        Processing events and send callbacks to handlers

        :param history: [list] list of :class:`Transaction`
        :param last_payment_id: id of last payment in history
        """
        history_iterator = iter(history[::-1])

        while self.offset < last_payment_id:
            try:
                payment = next(history_iterator)
                await self.dispatcher.process_event(payment)
                self.offset = payment.transaction_id
                self.offset_start_date = self.offset_end_date
            except StopIteration:  # handle exhausted iterator
                break

    def start_polling(
            self, *,
            get_updates_from: datetime = datetime.now(),
            timeout: Union[float, int, ClientTimeout] = DEFAULT_TIMEOUT,
            **kwargs
    ):
        try:
            self._loop.run_until_complete(self.welcome())
            self._loop.create_task(self._start_polling(
                get_updates_from=get_updates_from,
                timeout=timeout
            ))
            if isinstance(self.tg_app, BaseProxy):
                self.tg_app.setup(**kwargs, loop=self._loop)
            api_helper.run_forever_safe(loop=self._loop)
        except (SystemExit, KeyboardInterrupt):  # pragma: no cover
            # Allow to graceful shutdown
            pass
        finally:
            self._polling = False
            self._on_shutdown()

    def start_webhook(
            self, *,
            host: str = "localhost",
            port: int = 8080,
            path: Optional[Path] = None,
            app: Optional[web.Application] = None
    ):
        application = app or web.Application()

        hook_config, key = self._loop.run_until_complete(self.client.bind_webhook())

        server.setup(
            dispatcher=self.dispatcher,
            app=application,
            path=path,
            secret_key=self.client.secret_p2p,
            base64_key=key,
            tg_app=self.tg_app,
            host=host
        )

        try:
            self._loop.run_until_complete(self.welcome())
            web.run_app(application, host=host, port=port)
        except (KeyboardInterrupt, SystemExit):
            # Allow to graceful shutdown
            pass
        finally:
            self._on_shutdown()

    async def welcome(self) -> None:
        """ Execute on_startup callback"""
        self.dispatcher.logger.debug("Start polling!")
        for callback in self._on_startup_calls:
            await _inspect_and_execute_callback(
                callback=callback,
                client=self.client
            )

    async def goodbye(self) -> None:
        """ Execute on_shutdown callback """
        self.dispatcher.logger.debug("Goodbye!")
        for callback in self._on_shutdown_calls:
            await _inspect_and_execute_callback(
                callback=callback,
                client=self.client
            )



from __future__ import annotations

from asyncio import as_completed, set_event_loop_policy
from itertools import repeat
from typing import (
    AsyncGenerator
)
from typing import Dict, Optional, Any, Union, Tuple, Type, Iterable, cast, List

import aiohttp
from aiohttp import (
    ClientTimeout,
    ClientProxyConnectionError,
    ServerDisconnectedError,
    ClientConnectionError, ClientSession
)
from aiohttp.typedefs import LooseCookies

from glQiwiApi.core import AbstractParser
from glQiwiApi.core.constants import DEFAULT_TIMEOUT
from glQiwiApi.types import Response

_ProxyBasic = Union[str, Tuple[str, aiohttp.BasicAuth]]
_ProxyChain = Iterable[_ProxyBasic]
_ProxyType = Union[_ProxyChain, _ProxyBasic]


def _retrieve_basic(basic: _ProxyBasic) -> Dict[str, Any]:
    from aiohttp_socks.utils import parse_proxy_url  # type: ignore

    proxy_auth: Optional[aiohttp.BasicAuth] = None

    if isinstance(basic, str):
        proxy_url = basic
    else:
        proxy_url, proxy_auth = basic

    proxy_type, host, port, username, password = parse_proxy_url(proxy_url)
    if isinstance(proxy_auth, aiohttp.BasicAuth):
        username = proxy_auth.login
        password = proxy_auth.password

    return dict(
        proxy_type=proxy_type,
        host=host,
        port=port,
        username=username,
        password=password,
        rdns=True,
    )


def _prepare_connector(
        chain_or_plain: _ProxyType
) -> Tuple[Type["aiohttp.TCPConnector"], Dict[str, Any]]:
    from aiohttp_socks import ChainProxyConnector, ProxyConnector, ProxyInfo  # type: ignore

    # since tuple is Iterable(compatible with _ProxyChain) object, we assume that
    # user wants chained proxies if tuple is a pair of string(url) and BasicAuth
    if isinstance(chain_or_plain, str) or (
            isinstance(chain_or_plain, tuple) and len(chain_or_plain) == 2
    ):
        chain_or_plain = cast(_ProxyBasic, chain_or_plain)
        return ProxyConnector, _retrieve_basic(chain_or_plain)

    chain_or_plain = cast(_ProxyChain, chain_or_plain)
    infos: List[ProxyInfo] = []
    for basic in chain_or_plain:
        infos.append(ProxyInfo(**_retrieve_basic(basic)))

    return ChainProxyConnector, dict(proxy_infos=infos)


class HttpXParser(AbstractParser):
    """
    Aiohttp wrapper, implements the method of sending a request

    """

    _sleep_time = 2

    def __init__(self, proxy: Optional[_ProxyType] = None) -> None:
        self.base_headers = {
            'User-Agent': "glQiwiApi/1.0beta",
            'Accept-Language': "ru-RU,ru;q=0.9,en-US;q=0.8,en;q=0.7"
        }
        self._timeout = ClientTimeout(
            total=5,
            connect=None,
            sock_connect=5,
            sock_read=None
        )
        self._session: Optional[ClientSession] = None
        self._connector_type: Type[aiohttp.TCPConnector] = aiohttp.TCPConnector
        self._connector_init: Dict[str, Any] = {}
        self._should_reset_connector = False  # flag determines connector state
        self._proxy: Optional[_ProxyType] = None

        if proxy is not None:
            try:
                self._setup_proxy_connector(proxy)
            except ImportError as exc:  # pragma: no cover
                raise RuntimeError(
                    "In order to use aiohttp client for proxy requests, install "
                    "https://pypi.org/project/aiohttp-socks/"
                ) from exc

    def _setup_proxy_connector(self, proxy: _ProxyType) -> None:
        self._connector_type, self._connector_init = _prepare_connector(proxy)
        self._proxy = proxy

    @property
    def proxy(self) -> Optional[_ProxyType]:
        return self._proxy

    @proxy.setter
    def proxy(self, proxy: _ProxyType) -> None:
        self._setup_proxy_connector(proxy)
        self._should_reset_connector = True

    async def _make_request(
            self,
            url: str,
            get_json: bool = False,
            method: str = 'POST',
            set_timeout: bool = True,
            cookies: Optional[LooseCookies] = None,
            json: Optional[dict] = None,
            data: Optional[Dict[str, Union[
                str, int, List[
                    Union[str, int]
                ]]]
            ] = None,
            headers: Optional[dict] = None,
            params: Optional[
                Dict[str, Union[str, int, List[
                    Union[str, int]
                ]]]
            ] = None,
            get_bytes: bool = False,
            **kwargs) -> Response:
        """
        Send request to some url. Method has a similar signature with the `aiohttp.request`


        :param url: ссылка, куда вы хотите отправить ваш запрос
        :param get_json: указывает на то, хотите ли вы получить ответ
         в формате json
        :param method: Тип запроса
        :param data: payload data
        :param set_timeout:
        :param json:
        :param cookies: куки запроса
        :param headers: заголовки запроса
        :param params:
        :param get_bytes: указывает на то, хотите ли вы получить ответ
         в байтах
        :param kwargs:
        :return: Response instance
        """
        headers = headers or self.base_headers
        # Create new session if old was closed
        await self.create_session(
            timeout=self._timeout if set_timeout else DEFAULT_TIMEOUT
        )

        # sending query to some endpoint url
        try:
            response = await self._session.request(
                method=method,
                url=url,
                data=data,
                headers=headers,
                json=json if isinstance(json, dict) else None,
                cookies=cookies,
                params=params,
                **kwargs
            )
        except (
                ClientProxyConnectionError,
                ServerDisconnectedError,
                ClientConnectionError
        ):
            self.raise_exception(status_code='400_special_bad_proxy')
            return Response.bad_response()
        # Get content and return response
        if get_json:
            data = await response.json()
        elif get_bytes:
            data = await response.read()
        else:
            data = await response.text()

        return Response(
            status_code=response.status,
            response_data=data,
            raw_headers=response.raw_headers,
            cookies=response.cookies,
            ok=response.ok,
            content_type=response.content_type,
            host=response.host,
            url=response.url.__str__()
        )

    async def fetch(
            self,
            *,
            times: int = 1,
            **kwargs
    ) -> AsyncGenerator[Response, None]:
        """
        Basic usage: \n
        parser = HttpXParser() \n
        async for response in parser.fetch():
            print(response)

        :param times: quantity requests
        :param kwargs: HttpXParser._request kwargs
        :return:
        """
        coroutines = [self._make_request(**kwargs) for _ in repeat(None, times)]
        for future in as_completed(fs=coroutines):
            yield await future

    def fast(self) -> HttpXParser:
        """
        Method to fetching faster with using faster event loop(uvloop) \n
        USE IT ONLY ON LINUX SYSTEMS,
        on Windows or Mac its dont give performance!

        :return:
        """
        try:
            from uvloop import EventLoopPolicy  # type: ignore
            set_event_loop_policy(EventLoopPolicy())
        except ImportError:
            # Catching import error and forsake standard policy
            from asyncio import DefaultEventLoopPolicy as EventLoopPolicy  # type: ignore
            set_event_loop_policy(EventLoopPolicy())
        return self

    async def create_session(self, **kwargs) -> Optional[aiohttp.ClientSession]:
        """ Creating new session if old was close or it's None """
        if self.proxy is not None:
            kwargs.update(connector=self._connector_type(**self._connector_init))

        if self._should_reset_connector and isinstance(self._session, ClientSession):
            await self._session.close()

        if not isinstance(self._session, ClientSession):
            self._session = ClientSession(**kwargs)
            self._should_reset_connector = False
        elif isinstance(self._session, ClientSession):
            if self._session.closed:
                self._session = ClientSession(**kwargs)
                self._should_reset_connector = False
        return self._session

    async def close(self) -> None:
        """ close aiohttp session"""
        if isinstance(self._session, ClientSession):
            if not self._session.closed:
                await self._session.close()

    async def stream_content(
            self, url: str, timeout: int, chunk_size: int
    ) -> AsyncGenerator[bytes, None]:
        await self.create_session()

        async with self._session.get(url, timeout=timeout) as resp:
            async for chunk in resp.content.iter_chunked(chunk_size):
                yield chunk



from pydantic import validator

from glQiwiApi.types.base import Base
from glQiwiApi.types.qiwi_types.currency_parsed import CurrencyModel
from glQiwiApi.utils.currency_util import Currency


class Balance(Base):
    """ object: Balance """
    alias: str
    currency: CurrencyModel

    @validator("currency", pre=True, check_fields=True)
    def humanize_pay_currency(cls, v):
        if not isinstance(v, int):
            return v
        return Currency.get(str(v))



import glQiwiApi.types.qiwi_types as types
from glQiwiApi.core.web_hooks.filter import Filter

# Default filter for transaction handler
transaction_webhook_filter: Filter = Filter(
    lambda update: isinstance(update, (types.WebHook, types.Transaction))
)

# Default filter for bill handler
bill_webhook_filter: Filter = Filter(
    lambda update: isinstance(update, types.Notification)
)



from typing import Dict, Optional, Any, Union, NoReturn

import aiohttp

from glQiwiApi.core.basic_requests_api import HttpXParser, _ProxyType
from glQiwiApi.core.storage import Storage
from glQiwiApi.types import Response
from glQiwiApi.types.basics import Cached, DEFAULT_CACHE_TIME
from glQiwiApi.utils.exceptions import RequestError


class RequestManager(HttpXParser):
    """
    Deal with :class:`Storage`,
    caching queries and managing stable work of sending requests

    """
    __slots__ = (
        'without_context', 'messages', '_cache', '_should_reset_connector',
        '_connector_type', '_connector_init', '_proxy'
    )

    def __init__(
            self,
            without_context: bool = False,
            messages: Optional[Dict[str, str]] = None,
            cache_time: Union[float, int] = DEFAULT_CACHE_TIME,
            proxy: Optional[_ProxyType] = None
    ) -> None:
        super(RequestManager, self).__init__(proxy=proxy)

        self.without_context: bool = without_context
        self.messages: Optional[Dict[str, str]] = messages
        self._cache: Storage = Storage(cache_time=cache_time)

    def reset_cache(self) -> None:
        """ Clear all cache in storage """
        self._cache.clear(force=True)

    async def make_request(self, **kwargs) -> Response:
        """ The user-friendly method that allows sending requests to any URL  """
        return await super(RequestManager, self)._make_request(**kwargs)

    async def _make_request(self, *args, **kwargs) -> Response:
        """ Send request to service(API) """
        # Получаем текущий кэш используя ссылку как ключ
        response = self._cache[(kwargs.get('url'))]
        if not self._cache.validate(kwargs):
            try:
                response = await super()._make_request(*args, **kwargs)
            except aiohttp.ContentTypeError:
                raise RequestError(message="Unexpected error. Cannot deserialize answer.",
                                   status_code="unknown")

        # Проверяем, не был ли запрос в кэше, если нет,
        # то проверяем статус код и если он не 200 - выбрасываем ошибку
        if not isinstance(response, Cached):
            await self._close_session()
            if response.status_code != 200:
                self.raise_exception(
                    str(response.status_code),
                    json_info=response.response_data
                )

        self._cache_all(response, kwargs)

        return response

    async def _close_session(self):
        if self.without_context:
            await super(RequestManager, self).close()

    async def close(self) -> None:
        """ Close aiohttp session and reset cache data """
        await super(RequestManager, self).close()
        self.reset_cache()

    def raise_exception(
            self,
            status_code: str,
            json_info: Optional[Dict[str, Any]] = None,
            message: Optional[str] = None
    ) -> NoReturn:
        """ Raise RequestError exception with pretty explanation """
        if not isinstance(message, str):
            if self.messages is not None:
                message = self.messages.get(str(status_code), "Unknown")
        raise RequestError(
            message,
            status_code,
            additional_info=f"0.2.23 version api",
            json_info=json_info
        )

    def _cache_all(self, response: Response, kwargs: Dict[Any, Any]):
        resolved: Cached = self._cache.convert_to_cache(
            result=response.response_data,
            kwargs=kwargs,
            status_code=response.status_code
        )
        self._cache[kwargs["url"]] = resolved

    @property
    def is_session_closed(self) -> bool:
        if isinstance(self._session, aiohttp.ClientSession):
            if not self._session.closed:
                return True
        return False

    @classmethod
    def filter_dict(cls, dictionary: dict) -> dict:
        """
        Pop NoneType values and convert everything to str, designed?for=params
        :param dictionary: source dict
        :return: filtered dict
        """
        return {k: str(v) for k, v in dictionary.items() if v is not None}



from .types import (
    OperationType,
    Operation,
    OperationDetails,
    PreProcessPaymentResponse,
    Payment,
    AccountInfo,
    IncomingTransaction
)

__all__ = (
    'AccountInfo', 'OperationType', 'Operation',
    'OperationDetails',
    'PreProcessPaymentResponse', 'Payment', 'IncomingTransaction'
)



import json
from typing import Optional, Dict, Any, Union


class NoUrlFound(Exception):
    """Данная ошибка возникает при неправильной авторизации yoomoney"""


class RequestProxyError(Exception):
    """Возникает, если были переданы неправильные параметры запроса"""


ProxyError = Exception()


class InvalidCardNumber(Exception):
    """
    Ошибка, при передаче номера карты в неправильном формате

    """


class RequestAuthError(Exception):
    """
    Ошибка при неправильной аутентификации POST or GET data

    """


class InvalidToken(Exception):
    """
    Ошибка, возникающая, если был передан неверный токен

    """


class InvalidData(Exception):
    """
    Ошибка возникает если были переданы или получены невалидные данные

    """


class NoUpdatesToExecute(Exception):
    """
    Ошибка возбуждается если при полинге нет транзакций, чтобы обрабатывать

    """


class RequestError(Exception):
    """
    Возникает при ошибках сервиса или неправильной передаче параметров

    """

    def __init__(
            self,
            message: Optional[str],
            status_code: str,
            additional_info: Optional[str] = None,
            json_info: Optional[Union[Dict[str, Any], str]] = None
    ) -> None:
        super().__init__()
        self.message = message
        self.status_code = status_code
        self.additional_info = additional_info
        self.json_info = json_info

    def __str__(self) -> str:
        resp = "code={sc} doc={msg}, additional_info={info}"""
        return resp.format(
            sc=self.status_code,
            msg=self.message,
            info=self.additional_info
        )

    def __repr__(self) -> str:
        return self.__str__()

    def json(self) -> str:
        return json.dumps(
            self.__str__() if not self.json_info else self.json_info,
            indent=2,
            ensure_ascii=False
        )


__all__ = (
    'InvalidData',
    'NoUrlFound',
    'RequestAuthError',
    'RequestProxyError',
    'ProxyError',
    'InvalidCardNumber',
    'InvalidToken',
    'RequestError',
    'NoUpdatesToExecute'
)



import asyncio
import logging
from typing import List, Tuple, Coroutine, Any, Union

from .config import EventHandlerFunctor, EventFilter, E
from .filter import Filter
from ..builtin import (
    bill_webhook_filter,
    transaction_webhook_filter,
    InterceptHandler
)


def _setup_logger() -> logging.Logger:
    logger = logging.getLogger(__name__)
    logger.setLevel(level=logging.DEBUG)
    if not logger.handlers:
        logger.addHandler(InterceptHandler())
    return logger


class EventHandler:
    """
    Event handler, which executing and working with handlers

    """

    def __init__(self, functor: EventHandlerFunctor, *filter_: Tuple[Filter]) -> None:
        """

        :param functor:
        :param filter_:
        """
        self._fn = functor
        self._filters = filter_

    async def check_then_execute(self, event: E) -> Coroutine[Any, Any, Any]:
        """
        Check event, apply all filters and than pass on to handler

        :param event: handler
        """
        for filter_ in self._filters:
            if filter_.awaitable:
                if not await filter_.function(event):
                    break
            else:
                if not filter_.function(event):
                    break
        else:
            return await self._fn(event)


class Dispatcher:
    """
    Class, which managing all handlers

    """

    def __init__(self, loop: asyncio.AbstractEventLoop):
        if not isinstance(loop, asyncio.AbstractEventLoop):
            raise RuntimeError(
                f"Listener must have its event loop implemented with"
                f" {asyncio.AbstractEventLoop!r}"
            )

        self._loop = loop
        self.transaction_handlers: List[EventHandler] = []
        self.bill_handlers: List[EventHandler] = []
        self._logger: logging.Logger = _setup_logger()

    def register_transaction_handler(
            self,
            event_handler: EventHandlerFunctor,
            *filters: EventFilter
    ):
        self.transaction_handlers.append(
            self.wrap_handler(
                event_handler, filters,
                default_filter=transaction_webhook_filter
            )
        )

    def register_bill_handler(self, event_handler: EventHandlerFunctor,
                              *filters: EventFilter):
        self.bill_handlers.append(
            self.wrap_handler(
                event_handler, filters,
                default_filter=bill_webhook_filter
            )
        )

    @property
    def handlers(self) -> List[EventHandler]:
        """ Return all registered handlers """
        return [*self.bill_handlers, *self.transaction_handlers]

    @property
    def logger(self) -> logging.Logger:
        return self._logger

    @logger.setter
    def logger(self, logger: logging.Logger):
        self._logger = logger

    @staticmethod
    def wrap_handler(
            event_handler: EventHandlerFunctor,
            filters: Tuple[Union[EventFilter, Filter], ...],
            default_filter: Filter
    ) -> EventHandler:
        """
        Add new event handler.
        (!) Allows chain addition.
        :param event_handler: event handler, low order function
         which works with events
        :param filters: filter for low order function execution
        :param default_filter: default filter for handler
        :return: this handler manager
        """
        if filters:  # Initially filters are in tuple
            filters_list = list(filters)

            for index, filter_ in enumerate(filters):
                if not isinstance(filter_, Filter):
                    filters_list[index] = default_filter & Filter(filter_)

            filters = tuple(filters_list)
        else:
            filters = (default_filter,)

        return EventHandler(event_handler, *filters)

    def transaction_handler_wrapper(self, *filters: EventFilter):

        def decorator(callback: EventHandlerFunctor) -> EventHandlerFunctor:
            self.register_transaction_handler(callback, *filters)
            return callback

        return decorator

    def bill_handler_wrapper(self, *filters: EventFilter):

        def decorator(callback: EventHandlerFunctor) -> EventHandlerFunctor:
            self.register_bill_handler(callback, *filters)
            return callback

        return decorator

    async def process_event(self, event: E) -> None:
        """
        Feed handlers with event.
        :param event: any object that will be translated to handlers
        """

        for handler in self.handlers:
            await handler.check_then_execute(event)



import pathlib

import setuptools

PATH = pathlib.Path(__file__).parent

README = (PATH / "README.md").read_text()

REQUIREMENTS = (PATH / "SETUP_REQUIREMENTS.txt").read_text()

setuptools.setup(
    packages=setuptools.find_packages(exclude=(
        'tests', 'examples', 'examples.*', 'tests', 'tests.*')
    ),
    name="glQiwiApi",  # Replace with your own username
    version="1.0.0",
    author="GLEF1X",
    author_email="glebgar567@gmail.com",
    description="Light and fast wrapper of QIWI and YooMoney api's",
    package_data={"glQiwiApi": ["py.typed", "*.pyi", "**/*.pyi"]},
    # Длинное описание, которое будет отображаться на странице PyPi.
    # Использует README.md репозитория для заполнения.
    long_description=README,
    # Определяет тип контента, используемый в long_description.
    long_description_content_type="text/markdown",
    license="MIT",
    url="https://github.com/GLEF1X/glQiwiApi",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
    ],
    install_requires=REQUIREMENTS,
    python_requires=">=3.7",
)



import logging

from loguru import logger


class InterceptHandler(logging.Handler):

    def __init__(self, level=logging.NOTSET) -> None:
        super(InterceptHandler, self).__init__(level)

    def emit(self, record):
        # Get corresponding Loguru level if it exists
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno

        # Find caller from where originated the logged message
        frame, depth = logging.currentframe(), 2
        while frame.f_code.co_filename == logging.__file__:
            frame = frame.f_back
            depth += 1

        logger.opt(depth=depth, exception=record.exc_info).log(
            level,
            record.getMessage()
        )
